
//		FixNames.cpp - plugin to rename filenames containing troublesome characters


#define UNICODE						1
#define _UNICODE					1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR					0
#define VER_MINOR					45
#define VER_BUILD					1
#define VER_PATCH					0


#define PLUGIN_NAME					L"FixNames"


#ifdef _WIN64
#define PLATFORM					L"x64"
#else
#define PLATFORM					L"x86"
#endif

#define MAX_ARG_LEN					16384
#define SHORT_BUF_LEN				1024
#define MAX_ARGS					1024
#define MAX_FILENAME_LEN 16384

#define ERRHANDLE					GetStdHandle( STD_ERROR_HANDLE )

#define FLAG_PROMPT					0x00001
#define FLAG_NOT_REALLY				0x00002
#define FLAG_RENAME_DIRS			0x00004
#define FLAG_NO_FIX_COLLISIONS		0x00008
#define FLAG_EQUALS_HEX				0x00010
#define FLAG_HTML_ENTITIES			0x00020
#define FLAG_NO_PERCENTS			0x00040
#define FLAG_REPLACE_BLANKS			0x00080
#define FLAG_REPLACE_AMPS			0x00100
#define FLAG_REPLACE_PARENTHESES	0x00200
#define FLAG_REPLACE_HIGH_CHARS		0x00400
#define FLAG_NO_SUMMARY				0x00800
#define FLAG_NO_ERROR_MESSAGES		0x01000
#define FLAG_RECURSE				0x02000
#define FLAG_NO_JUNCTIONS			0x04000
#define FLAG_NO_HIDDEN_DIRS			0x08000
#define FLAG_CONDENSE_WHITESPACE	0x10000
#define FLAG_CONDENSE_PUNCTUATION	0x20000
#define FLAG_STRIP_ALL_QUOTES		0x40000

#define SPACE						0x20
#define QUOTE						0x22
#define NBSPACE						0xa0

#define USER_ABORT					-3



struct Totals_Struct {
	unsigned long int Found;
	unsigned long int Fixed;
	unsigned long int Skipped;
	unsigned long int Failed;
};


const wchar_t PluginFeatures[] = L"fixnames";


const wchar_t HelpWindowTitle[] = L"FixNames Plugin Help";


const wchar_t FixNamesHelpText[] = L"Rename files to remove troublesome characters.  \n\
\n\
FIXNAMES /A /B /C /D /E /H /K: /N /P /Q:list /R:list /S /U filespec...  \r\n\n\
\t/A\treplace ampersands with underscores\n\
\t/B\treplace blanks with underscores\n\
\t/C\treplace parentheses with underscores\n\
\t/D\trename directories, not files\n\
\t/E\tdecode =XX hex pairs\n\
\t/H\tdecode HTML entities\n\
\t/K:\tcondense repeated Blanks/Punctuation\n\
\t/N\tnot really\n\
\t/P\tprompt before each\n\
\t/Q:\tlist of characters to replace\n\
\t/R:\treplacement characters; use with /Q:\n\
\t/S\trecurse into Subdirectories\n\
\t/U\treplace high-order Unicode characters\n\
\n\
Use wildcards to match the difficult characters.  Note that you  \n\
do not supply a \u201cnew\u201d or replacement filename; the plugin\n\
generates its own.  These new names can then be changed with  \n\
REN, if desired.\n";


int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



#include "ParseArgs.cpp"



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



/*
int IsIllegalFilename( LPTSTR Filename )
{
	if ( Filename == NULL )
		return 1024;
	else if ( Filename[0] == '\0' )
		return 4;

	int i = 0;
	wchar_t ch = '\0';
	int rv = 0;

	while ( Filename[i] ) {
		ch = Filename[i++];

		if ( ( ch == '/' ) || ( ch == '?' ) || ( ch == '<' ) || ( ch == '>' ) || ( ch == '\\' ) || ( ch == ':' ) || ( ch == '*' ) || ( ch == '|' ) || ( ch == QUOTE ) )
			rv |= 1;
	}

	if ( ( ch == SPACE ) || ( ch == '.' ) )
		rv |= 2;

	return rv;
}
*/



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		Sprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



#include "HtmlEntities.cpp"



LPTSTR _PathPart( LPTSTR Filename, LPTSTR outPath )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 )
		return NULL;

	if ( outPath ) {
		for ( size_t i = 0; i <= EndOfPathPart; i++ )
			outPath[i] = Filename[i];

		outPath[EndOfPathPart + 1] = '\0';
		return outPath;
	}

	return Filename;
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



int _AddWildcardIfDirectory( LPTSTR Filename, size_t BufferSize )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	if ( QueryIsDirectory( Filename ) )
		AppendFilename( Filename, L"*", BufferSize );

	return 0;
}



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



int DisplayTotals( Totals_Struct Totals )
{
	wchar_t FoundStr[SHORT_BUF_LEN]   = L"";
	wchar_t FixedStr[SHORT_BUF_LEN]   = L"";
	wchar_t FailedStr[SHORT_BUF_LEN]  = L"";
	wchar_t SkippedStr[SHORT_BUF_LEN] = L"";

	wsprintf( FoundStr,   L"%u", Totals.Found );
	wsprintf( FixedStr,   L"%u", Totals.Fixed );
	wsprintf( FailedStr,  L"%u", Totals.Failed );
	wsprintf( SkippedStr, L"%u", Totals.Skipped );

	InsertCommas( FoundStr, SHORT_BUF_LEN );
	InsertCommas( FixedStr, SHORT_BUF_LEN );
	InsertCommas( FailedStr, SHORT_BUF_LEN );
	InsertCommas( SkippedStr, SHORT_BUF_LEN );

	Printf( L"Found %s, fixed %s, failed %s, skipped %s", FoundStr, FixedStr, FailedStr, SkippedStr );
	return 0;
}



wchar_t ReplaceChar( wchar_t ch )
{
	if ( ch == '\0' )
		return ch;
	else if ( ch < SPACE )
		return '_';
	else if ( ( ch == '%' ) || ( ch == '^' ) || ( ch == '/' ) || ( ch == '\\' ) || ( ch == '<' ) || ( ch == '>' ) || ( ch == '?' ) || ( ch == '*' ) || ( ch == ':' ) || ( ch == '|' ) )
		return '_';
	else if ( ( ch == '`' ) || ( ch == QUOTE ) )
		return 0x27;
	else if ( ( ch >= 0x80 ) && ( ch <= 0x9f ) ) {
		ch = Windows1252[ch - 0x80];

		if ( ( ch >= 0x80 ) && ( ch <= 0x9f ) )
			return '_';
		else
			return ch;
	}
	else if ( ch == NBSPACE )
		return SPACE;
	else if ( ( ch >= 0x2000 ) && ( ch <= 0x200a ) )
		return SPACE;
	else if ( ch == 0x205f )
		return SPACE;
	else if ( ( ch == 0x200b ) || ( ch == 0x202f ) || ( ch == 0xfeff ) )
		return '_';

	return ch;
}



wchar_t ReplaceChar2( wchar_t ch, LPCTSTR FromChar, LPCTSTR ToChar )
{
	if ( !ch || !FromChar )
		return ch;


	size_t i			= 0;


	if ( !wcschr( FromChar, towupper( ch ) ) )
		return ch;

	i = wcschr( FromChar, towupper( ch ) ) - FromChar;
	ch = '\0';

	if ( ToChar )
		if ( i < wcslen( ToChar ) )
			ch = ToChar[i];

	return ch;
}



int CondenseChars( LPTSTR String, unsigned int Flags )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;


	wchar_t temp[MAX_FILENAME_LEN]			= L"";

	wchar_t *ch			= String;
	wchar_t sv			= '\0';
	size_t j			= 0;
	int k				= 0;


	while ( *ch ) {
		if ( j < ( MAX_FILENAME_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';
		}

		if ( iswspace( *ch ) && ( Flags & FLAG_CONDENSE_WHITESPACE ) ) {
			if ( ( *ch != SPACE ) && iswspace( ch[1] ) )
				*ch = SPACE, k++;

			while ( ch[1] && iswspace( ch[1] ) )
				ch++, k++;
		}
		else if ( iswpunct( *ch ) && ( Flags & FLAG_CONDENSE_PUNCTUATION ) ) {
			sv = *ch;

			while ( ch[1] == sv )
				ch++, k++;
		}

		ch++;
	}

	if ( k )
		wcscpy_s( String, MAX_FILENAME_LEN, temp );

	return k;
}



int TryRenameUNC( LPTSTR OldFullName, LPTSTR NewName )
{
	if ( !OldFullName || !NewName )
		return -1;

	if ( NewName[0] == '\0' )
		return -1;

	if ( !iswalpha( OldFullName[0] ) )
		return 1;

	if ( ( OldFullName[1] != ':' ) || ( OldFullName[2] != '\\' ) || ( OldFullName[3] == '\0' ) )
		return 1;

	wchar_t OldNameUNC[MAX_FILENAME_LEN] = L"\\\\?\\";
	wcscat_s( OldNameUNC, MAX_FILENAME_LEN, OldFullName );

	if ( MoveFile( OldNameUNC, NewName ) )
		return 0;
	else
		return 1;
}



int TryRenameAltname( LPTSTR Directory, LPTSTR OldNameAlt, LPTSTR NewName )
{
	if ( !Directory || !OldNameAlt || !NewName )
		return -1;

	if ( ( Directory[0] == '\0' ) || ( OldNameAlt[0] == '\0' ) || ( NewName[0] == '\0' ) )
		return 1;

	wchar_t OldFullName[MAX_FILENAME_LEN] = L"";

	if ( iswalpha( Directory[0] ) )
		if ( ( Directory[1] == ':' ) && ( Directory[2] == '\\' ) )
			wcscpy_s( OldFullName, MAX_FILENAME_LEN, L"\\\\?\\" );

	wcscat_s( OldFullName, MAX_FILENAME_LEN, Directory );
	wcscat_s( OldFullName, MAX_FILENAME_LEN, OldNameAlt );

	if ( MoveFile( OldFullName, NewName ) )
		return 0;
	else
		return 1;
}



int ReplaceHighChars( LPTSTR String, size_t buflen )
{
	if ( !String || ( buflen < 64 ) )
		return -1;


	wchar_t *temp			= NULL;
	size_t j				= 0;
	int k					= 0;


	temp = (wchar_t*) malloc( buflen * sizeof( wchar_t ) );
	if ( !temp )
		return -2;

	for ( wchar_t *ch = String; *ch; ch++ ) {

		if ( ( *ch >= 0xd800 ) && ( *ch <= 0xdfff ) ) {
			temp[j++] = '_';
			k++;

			if ( ( *ch <= 0xdbff ) && ( ch[1] >= 0xdc00 ) && ( ch[1] <= 0xdfff ) )
				ch++;
		}
		else
			temp[j++] = *ch;

		if ( j >= buflen )
			return -3;

		temp[j] = '\0';
	}

	if ( k )
		wcscpy_s( String, buflen, temp );

	free( temp );
	return k;
}



int hdval( wchar_t ch )
{
	if ( ( ch >= '0' ) && ( ch <= '9' ) )
		return ch - '0';
	else if ( ( ch >= 'A' ) && ( ch <= 'F' ) )
		return ch - 'A' + 10;
	else if ( ( ch >= 'a' ) && ( ch <= 'f' ) )
		return ch - 'a' + 10;
	else
		return -1;
}



int MakeNewName( LPTSTR Directory, LPTSTR OldName, LPTSTR NewName, DWORD *Flags, LPCTSTR FromChar, LPCTSTR ToChar )
{
	wchar_t OldName2[MAX_FILENAME_LEN]		= L"";
	wchar_t Working[MAX_FILENAME_LEN]		= L"";
	wchar_t FullName[MAX_FILENAME_LEN]		= L"";
	wchar_t Ext[MAX_FILENAME_LEN]			= L"";
	wchar_t Number[SHORT_BUF_LEN]			= L"";

	wchar_t ch = '\0';

	int j = 0;
	unsigned long int nnn = 0;
	wchar_t *sep1 = NULL;


	wcscpy_s( OldName2, MAX_FILENAME_LEN, OldName );

	if ( *Flags & FLAG_HTML_ENTITIES )
		if ( wcschr( OldName, '&' ) && wcschr( OldName, ';' ) )
			Deentitize( OldName2, MAX_FILENAME_LEN );


	if ( *Flags & FLAG_REPLACE_HIGH_CHARS )
		ReplaceHighChars( OldName2, MAX_FILENAME_LEN );


	for ( size_t i = 0; OldName2[i]; i++ ) {
		ch = OldName2[i];

		if ( ( ( ( *Flags & FLAG_NO_PERCENTS ) == 0 ) && ( ch == '%' ) ) || ( ( *Flags & FLAG_EQUALS_HEX ) && ( ch == '=' ) ) ) {
			if ( iswxdigit( OldName2[i + 1] ) && iswxdigit( OldName2[i + 2] ) ) {
				ch = ( hdval( OldName2[i + 1] ) << 4 ) | hdval( OldName2[i + 2] );
				i += 2;
			}
		}

		if ( *Flags & FLAG_STRIP_ALL_QUOTES )
			if ( ( ch == '\'' ) || ( ch == '"' ) || ( ch == '`' ) )
				ch = '_';

		if ( FromChar )
			 ch = ReplaceChar2( ch, FromChar, ToChar );

		ch = ReplaceChar( ch );

		if ( *Flags & FLAG_REPLACE_BLANKS )
			if ( iswspace( ch ) )
				ch = '_';

		if ( *Flags & FLAG_REPLACE_AMPS )
			if ( ch == '&' )
				ch = '_';

		if ( *Flags & FLAG_REPLACE_PARENTHESES )
			if ( ( ch == '(' ) || ( ch == ')' ) )
				ch = '_';

		if ( ch && ( j < ( MAX_FILENAME_LEN - 1 ) ) ) {
			Working[j++] = ch;
			Working[j] = '\0';
		}
	};


	//		condense whitespace or punctuation as desired :

	if ( *Flags | ( FLAG_CONDENSE_WHITESPACE | FLAG_CONDENSE_PUNCTUATION ) )
		CondenseChars( Working, *Flags );


	//		check for spaces or periods at end of filename; remove any found:

	int l = (int) wcslen( Working );
	do {
		if ( l == 0 )
			break;
		else if ( ( Working[l - 1] != SPACE ) && ( Working[l - 1] != NBSPACE ) && ( Working[l - 1] != '.' ) )
			break;

		l--;
		Working[l] = '\0';
	} while ( TRUE );

	//		if the filename is empty, supply a stupid default:

	if ( Working[0] == '\0' ) {
		if ( *Flags & FLAG_RENAME_DIRS )
			wcscpy_s( Working, MAX_FILENAME_LEN, L"Dir" );
		else
			wcscpy_s( Working, MAX_FILENAME_LEN, L"File" );
	}

	//		exit if no changes:

	if ( !_wcsicmp( OldName, Working ) )
		return 1;

	if ( *Flags & FLAG_NO_FIX_COLLISIONS ) {
		wcscpy_s( NewName, MAX_FILENAME_LEN, Working );
		return 0;
	}

	wcscpy_s( FullName, MAX_FILENAME_LEN, Directory );
	wcscat_s( FullName, MAX_FILENAME_LEN, Working );
	
	if ( !QueryIsFileOrDirectory( FullName ) ) {
		wcscpy_s( NewName, MAX_FILENAME_LEN, Working );
		return 0;
	}

	//  Problem:  The new name already exists!

	ch = SPACE;
	Ext[0] = '\0';


	sep1 = wcsrchr( Working, '.' );
	if ( sep1 )
		wcscpy_s( Ext, MAX_FILENAME_LEN, sep1 );

	//	Printf( L"   * Working = \"%s\"\r\n", Working );
	//	Printf( L"   * Ext     = \"%s\"\r\n", Ext );

	nnn = 0;
	j = (int) wcslen( Working ) - 1;

	if ( iswdigit( Working[j] ) ) {

		while ( ( j > 0 ) && ( iswdigit( Working[j] ) ) )
			j--;

		if ( Working[j] == '-' ) {
			nnn = wcstoul( Working + j + 1, &sep1, 10 );
			if ( nnn > 9999 )
				nnn = 0;

			Working[j] = '\0';
		}
	}

	do {
		nnn++;
		swprintf( Number, SHORT_BUF_LEN, L"-%03u", nnn );
		wcscpy_s( FullName, MAX_FILENAME_LEN, Directory );
		wcscat_s( FullName, MAX_FILENAME_LEN, Working );
		wcscat_s( FullName, MAX_FILENAME_LEN, Number );
		wcscat_s( FullName, MAX_FILENAME_LEN, Ext );
	}
	while ( QueryIsFileOrDirectory( FullName ) );

	wcscpy_s( NewName, MAX_FILENAME_LEN, Working );
	wcscat_s( NewName, MAX_FILENAME_LEN, Number );
	wcscat_s( NewName, MAX_FILENAME_LEN, Ext );
	return 0;
}



int GetYNAQ( void )
{
	int key = 0;

	do {

		try {
			key = GetKeystroke( 0x0230 );
		}
		catch ( ... ) {
			return USER_ABORT;
		}

		if ( ( key == 'N' ) || ( key == 0x1b ) ) {
			QPuts( L"No  \n" );
			return 0;
		}
		else if ( key == 'Y' ) {
			_cputws( L"Yes \n" );
			return 1;
		}
		else if ( key == 'A' ) {
			_cputws( L"All \n" );
			return 2;
		}
		else if ( key == 'Q' ) {
			_cputws( L"Quit\n" );
			return USER_ABORT;
		}

	} while ( TRUE );

	return 0;
}



int RenameItem( LPTSTR Directory, LPTSTR OldName, LPTSTR OldNameAlt, DWORD *Flags, LPCTSTR FromChar, LPCTSTR ToChar, Totals_Struct *Totals )
{
	wchar_t NewName[MAX_FILENAME_LEN] = L"";

	if ( MakeNewName( Directory, OldName, NewName, Flags, FromChar, ToChar ) == 1 ) {
		(*Totals).Skipped++;
		return 0;
	}


	int key = 0;
	wchar_t OldFullName[MAX_FILENAME_LEN]	= L"";
	wchar_t NewFullName[MAX_FILENAME_LEN]	= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	DWORD LastError		= 0;
	int rv				= 0;


	wcscpy_s( OldFullName, MAX_FILENAME_LEN, Directory );
	wcscat_s( OldFullName, MAX_FILENAME_LEN, OldName );

	wcscpy_s( NewFullName, MAX_FILENAME_LEN, Directory );
	wcscat_s( NewFullName, MAX_FILENAME_LEN, NewName );


	if ( *Flags & FLAG_PROMPT ) {
		_cwprintf( L"\"%s\" \u2500\u25ba \"%s\"  (Y/N/A)? ", OldFullName, NewFullName );
		key = GetYNAQ();

		if ( key == USER_ABORT ) {
			(*Totals).Skipped++;
			return USER_ABORT;
		}
		else if ( key == 0 ) {
			(*Totals).Skipped++;
			return 0;
		}
		else if ( key == 2 )
			*Flags &= ( ~FLAG_PROMPT );
	}
	else
		Printf( L"\"%s\" \u2500\u25ba \"%s\"\r\n", OldFullName, NewFullName );


	if ( *Flags & FLAG_NOT_REALLY ) {
		(*Totals).Skipped++;
		return 0;
	}

	
	rv = MoveFile( OldFullName, NewFullName );
	if ( rv ) {
		(*Totals).Fixed++;
		return 0;
	}
	else
		LastError = GetLastError();

	if ( TryRenameUNC( OldFullName, NewName ) == 0 ) {
		(*Totals).Fixed++;
		return 0;
	}

	if ( OldNameAlt != NULL ) {
		if ( OldNameAlt[0] ) {
			if ( _wcsicmp( OldName, OldNameAlt ) ){
				if ( TryRenameAltname( Directory, OldNameAlt, NewName ) == 0 ) {
					(*Totals).Fixed++;
					return 0;
				}
			}
		}
	}

	/*
	if ( TryRenameUgly( Directory, OldName, NewName ) == 0 ) {
		(*Totals).Fixed++;
		return 0;
	}
	*/

	GetErrorText( LastError, ErrorText );
	if ( ( *Flags & FLAG_NO_ERROR_MESSAGES ) == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error renaming \"%s\":\r\n", OldName );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );
		Qprintf( ERRHANDLE, L"\r\n" );
	}

	(*Totals).Failed++;
	return 2;
}



int FindMatchingItems( LPTSTR Directory, LPTSTR Wildspec, DWORD *Flags, LPCTSTR FromChar, LPCTSTR ToChar, Totals_Struct *GrandTotal )
{
	if ( !Directory || !Wildspec )
		return -1;
	
	if ( ( Directory[0] == '\0' ) || ( Wildspec[0] == '\0' ) )
		return -1;

	wchar_t SearchString[MAX_FILENAME_LEN] = L"";
	Totals_Struct Totals = { 0, 0, 0, 0 };

	WIN32_FIND_DATA FoundItem;

	wcscpy_s( SearchString, MAX_FILENAME_LEN, Directory );
	wcscat_s( SearchString, MAX_FILENAME_LEN, L"*" );

	int rirv		= 0;
	int fmirv		= 0;
	int rv			= 0;

	HANDLE SearchHandle = FindFirstFile( SearchString, &FoundItem );

	if ( SearchHandle != INVALID_HANDLE_VALUE ) {

		do {
			if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != ( ( *Flags & FLAG_RENAME_DIRS ) ? FILE_ATTRIBUTE_DIRECTORY : 0 ) )
				continue;

			if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
				continue;

			if ( WildcardComparison( Wildspec, FoundItem.cFileName, 1, 1 ) )
				continue;

			Totals.Found++;
			rirv = RenameItem( Directory, FoundItem.cFileName, FoundItem.cAlternateFileName, Flags, FromChar, ToChar, &Totals );
			if ( rirv == USER_ABORT ) {
				rv = USER_ABORT;
				break;
			}
			else if ( rirv )
				rv = 2;

		} while ( FindNextFile( SearchHandle, &FoundItem ) );

		FindClose( SearchHandle );
	}

	(*GrandTotal).Found   += Totals.Found;
	(*GrandTotal).Fixed   += Totals.Fixed;
	(*GrandTotal).Failed  += Totals.Failed;
	(*GrandTotal).Skipped += Totals.Skipped;

	if ( rv == USER_ABORT )
		return rv;


	if ( Totals.Found == 0 ) {
		if ( ( *Flags & FLAG_NO_ERROR_MESSAGES ) == 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File not found:  \"%s%s\"\r\n", Directory, Wildspec );
			Qprintf( ERRHANDLE, L"\r\n" );
		}
	}
	else if ( ( *Flags & FLAG_NO_SUMMARY ) == 0 ) {
		CrLf();
		DisplayTotals( Totals );
		Printf( L" in \"%s\"\r\n\n", Directory );
	}


	if ( *Flags & FLAG_RECURSE ) {
		SearchHandle = FindFirstFile( SearchString, &FoundItem );

		if ( SearchHandle != INVALID_HANDLE_VALUE ) {

			do {
				if ( ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
					continue;

				if ( ( *Flags & FLAG_NO_JUNCTIONS ) && ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) )
					continue;

				if ( ( *Flags & FLAG_NO_HIDDEN_DIRS ) && ( FoundItem.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN ) )
					continue;

				if ( !wcscmp( FoundItem.cFileName, L"." ) || !wcscmp( FoundItem.cFileName, L".." ) )
					continue;

				wcscpy_s( SearchString, MAX_FILENAME_LEN, Directory );
				wcscat_s( SearchString, MAX_FILENAME_LEN, FoundItem.cFileName );
				wcscat_s( SearchString, MAX_FILENAME_LEN, L"\\" );

				fmirv = FindMatchingItems( SearchString, Wildspec, Flags, FromChar, ToChar, GrandTotal );

				if ( fmirv == USER_ABORT ) {
					rv = USER_ABORT;
					break;
				}
				else if ( fmirv )
					rv = 2;

			} while ( FindNextFile( SearchHandle, &FoundItem ) );

			FindClose( SearchHandle );
		}
	}

	return rv;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	if ( !HelpText )
		return 1;

	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s version %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI fixnames( LPTSTR inString )
{
	if ( !inString )
		return 1;

	DWORD Flags = 0;
	BOOL FoundFilename  = FALSE;

	wchar_t Filename[MAX_FILENAME_LEN]  = L"";
	wchar_t Directory[MAX_FILENAME_LEN] = L"";
	wchar_t Wildspec[MAX_FILENAME_LEN]  = L"";

	LPTSTR FromChar						= NULL;
	LPTSTR ToChar						= NULL;

	Totals_Struct GrandTotal = { 0, 0, 0, 0 };
	
	int rv  = 0, fmirv = 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( FixNamesHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( FixNamesHelpText, TRUE );
			else if ( !wcscmp( arg[i], L"/A" ) )
				Flags |= FLAG_REPLACE_AMPS;
			else if ( !wcscmp( arg[i], L"/B" ) )
				Flags |= FLAG_REPLACE_BLANKS;
			else if ( !wcscmp( arg[i], L"/C" ) )
				Flags |= FLAG_REPLACE_PARENTHESES;
			else if ( !wcscmp( arg[i], L"/D" ) )
				Flags |= FLAG_RENAME_DIRS;
			else if ( !wcscmp( arg[i], L"/E" ) )
				Flags |= FLAG_EQUALS_HEX;
			else if ( !wcscmp( arg[i], L"/F" ) )
				Flags |= FLAG_STRIP_ALL_QUOTES;
			else if ( !wcscmp( arg[i], L"/H" ) )
				Flags |= FLAG_HTML_ENTITIES;
			else if ( !wcscmp( arg[i], L"/K" ) )
				Flags |= ( FLAG_CONDENSE_WHITESPACE | FLAG_CONDENSE_PUNCTUATION );
			else if ( !wcsncmp( arg[i], L"/K", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				while ( arg[i][j] ) {
					switch ( arg[i][j] ) {

						case 'B' :
							Flags |= FLAG_CONDENSE_WHITESPACE;
							break;

						case 'P' :
							Flags |= FLAG_CONDENSE_PUNCTUATION;
							break;

						default :
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown suboption: /K:%c\r\n", arg[i][j] );
							return 1;
					}
					
					j++;
				}
			}
			else if ( !wcscmp( arg[i], L"/N" ) )
				Flags |= FLAG_NOT_REALLY;
			else if ( !wcsncmp( arg[i], L"/N", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				while ( arg[i][j] ) {
					switch ( arg[i][j] ) {

						case 'D' :
							Flags |= FLAG_NO_HIDDEN_DIRS;
							break;

						case 'E' :
							Flags |= FLAG_NO_ERROR_MESSAGES;
							break;

						case 'F' :
							Flags |= FLAG_NO_FIX_COLLISIONS;
							break;

						case 'J' :
							Flags |= FLAG_NO_JUNCTIONS;
							break;

						case 'P' :
							Flags |= FLAG_NO_PERCENTS;
							break;

						case 'S' :
							Flags |= FLAG_NO_SUMMARY;
							break;

						default :
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L" Unknown suboption /N:%c\r\n", arg[i][j] );
							return 1;
					}

					j++;
				}
			}
			else if ( !wcscmp( arg[i], L"/P" ) )
				Flags |= FLAG_PROMPT;
			else if ( !wcsncmp( arg[i], L"/Q", 2 ) ) {
				FromChar = arg[i] + 2;
				if ( IsColonOrEquals( FromChar[0] ) )
					FromChar++;

				if ( FromChar[0] == '\0' )
					FromChar = NULL;
			}
			else if ( !wcsncmp( arg[i], L"/R", 2 ) ) {
				ToChar = arg[i] + 2;
				if ( IsColonOrEquals( ToChar[0] ) )
					ToChar++;

				if ( ToChar[0] == '\0' )
					ToChar = NULL;
			}
			else if ( !wcscmp( arg[i], L"/S" ) )
				Flags |= FLAG_RECURSE;
			else if ( !wcscmp( arg[i], L"/U" ) )
				Flags |= FLAG_REPLACE_HIGH_CHARS;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else
				FoundFilename = TRUE;

		}
	}

	if ( !FoundFilename ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filespec\r\n" );
		return -1;
	}

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( arg[i][0] != '/' ) {
				QueryTrueName( arg[i], Filename );

				if ( ( Flags & FLAG_RENAME_DIRS ) == 0 )
					_AddWildcardIfDirectory( Filename, MAX_FILENAME_LEN );

				_PathPart( Filename, Directory );
				_FilenamePart( Filename, Wildspec );

				fmirv = FindMatchingItems( Directory, Wildspec, &Flags, FromChar, ToChar, &GrandTotal );

				if ( fmirv == USER_ABORT )
					return 3;
				else if ( fmirv )
					rv = 2;
			}
		}
	}

	if ( ( Flags & FLAG_RECURSE ) && ( ( Flags & FLAG_NO_SUMMARY ) == 0 ) ) {
		QPuts( L"GRAND TOTAL: " );
		DisplayTotals( GrandTotal );
		CrLf();
	}

	return rv;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll = (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor = (LPTSTR) L"Charles Dye";
	piInfo.pszEmail = (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW = (LPTSTR) L"http://charlesdye.net/plugins/fixnames.html";
	piInfo.pszDescription = (LPTSTR) L"Renames filenames containing troublesome characters";
	piInfo.pszFunctions = (LPTSTR) PluginFeatures;
	piInfo.nMajor = VER_MAJOR;
	piInfo.nMinor = VER_MINOR;
	piInfo.nBuild = VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t helpbuf[MAX_ARG_LEN] = L"";

	if ( !_wcsicmp( inString, L"fixnames" ) ) {
		wcscpy_s( helpbuf, MAX_ARG_LEN, FixNamesHelpText );
		MessageBox( GetForegroundWindow(), helpbuf, HelpWindowTitle, MB_OK );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"fixnames" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( FixNamesHelpText ) );
		return 1;
	}
	else
		return 0;
}
